/*$
Copyright (c) 2017, Azel
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
$*/

/**********************************
 * main
 **********************************/

#include <stdio.h>

#include "mDef.h"
#include "mStr.h"
#include "mGui.h"
#include "mEvent.h"
#include "mSysCol.h"
#include "mTrans.h"

#include "mWidget.h"
#include "mContainer.h"
#include "mWindow.h"
#include "mWindowDef.h"

#include "mMenuBar.h"
#include "mLabel.h"
#include "mListView.h"
#include "mButton.h"
#include "mCheckButton.h"
#include "mColorButton.h"
#include "mScrollBar.h"
#include "mIconButtons.h"
#include "mGroupBox.h"
#include "mComboBox.h"
#include "mLineEdit.h"
#include "mTab.h"
#include "mSliderBar.h"
#include "mProgressBar.h"
#include "mDockWidget.h"

#include "mSysDialog.h"

#include "mMenu.h"
#include "mAccelerator.h"
#include "mPixbuf.h"
#include "mFont.h"
#include "mImageList.h"
#include "mIniWrite.h"

#include "def.h"

#include "deftrans.h"


//-------------------

#define LISTITEM_COLORBOX_W   20
#define LISTITEM_COLORBOX_SEP 2

#define _ABOUT_TEXT  "mThemeEditor ver 1.0.0\n\nCopyright (C) 2017 Azel"

//-------------------


//===========================
// sub
//===========================


/** ファイル読み込み */

static void _load_file(MainWindow *p,const char *filename)
{
	mAppLoadThemeFile(filename);
	
	mColorButtonSetColor(p->colbtt, SYSCOL_RGB(p->selitem->param));
}

/** ファイルに書き込み */

static void _save_file(const char *filename)
{
	FILE *fp;
	int i;

	fp = mIniWriteOpenFile(filename);
	if(!fp) return;

	mIniWriteGroup(fp, "color");

	for(i = 0; i < MSYSCOL_NUM - 1; i++)
		mIniWriteHex(fp, g_color_name[i], SYSCOL_RGB(i));

	fclose(fp);
}

/** 開く */

static void _cmd_open_file(MainWindow *p)
{
	mStr str = MSTR_INIT;

	if(!mSysDlgOpenFile(M_WINDOW(p), "All files (*)\t*", 0, p->strOpenDir.buf, 0, &str))
		return;

	mStrPathGetDir(&p->strOpenDir, str.buf);

	mAppLoadThemeFile(str.buf);
	mColorButtonSetColor(p->colbtt, SYSCOL_RGB(p->selitem->param));

	mStrFree(&str);
}

/** 保存 */

static void _cmd_save_file(MainWindow *p)
{
	mStr str = MSTR_INIT;
	char *dir;

	if(!mStrIsEmpty(&p->strSaveDir))
		dir = p->strSaveDir.buf;
	else if(!mStrIsEmpty(&p->strOpenDir))
		dir = p->strOpenDir.buf;
	else
		dir = NULL;

	if(!mSysDlgSaveFile(M_WINDOW(p), "Theme file (*.theme)\t*.theme", 0, dir, 0, &str, NULL))
		return;

	mStrPathGetDir(&p->strSaveDir, str.buf);

	mStrPathSetExt(&str, "theme");

	_save_file(str.buf);

	mStrFree(&str);
}

/** 色変更時 */

static void _change_color(MainWindow *p,int col)
{
	int no = p->selitem->param;
	
	SYSCOL_RGB(no) = col;
	SYSCOL_PIX(no) = mRGBtoPix(col);

	mGuiUpdateAllWindows();
}


//===========================
// ハンドラ
//===========================


/** コマンド */

static void _event_command(MainWindow *p,int id)
{
	switch(id)
	{
		//開く
		case MENUID_FILE_OPEN:
			_cmd_open_file(p);
			break;
		//保存
		case MENUID_FILE_SAVE:
			_cmd_save_file(p);
			break;
		//リセット
		case MENUID_FILE_RESET:
			mAppLoadThemeFile(NULL);

			mColorButtonSetColor(p->colbtt, SYSCOL_RGB(p->selitem->param));
			break;
	
		//終了
		case MENUID_FILE_EXIT:
			mAppQuit();
			break;

		//about
		case MENUID_HELP_ABOUT:
			mSysDlgAbout(M_WINDOW(p), _ABOUT_TEXT);
			break;
	}
}

/** 通知 */

static void _event_notify(MainWindow *p,mEvent *ev)
{
	int type = ev->notify.type;

	switch(ev->notify.id)
	{
		//リスト
		case WID_LIST_COLOR:
			if(type == MLISTVIEW_N_CHANGE_FOCUS)
			{
				//リスト選択変更

				p->selitem = (mListViewItem *)ev->notify.param1;
				
				mColorButtonSetColor(p->colbtt, SYSCOL_RGB(ev->notify.param2));
			}
			else if(type == MLISTVIEW_N_ITEM_DBLCLK)
			{
				//ダブルクリック -> 色選択

				mRgbCol col;

				col = SYSCOL_RGB(p->selitem->param);

				if(mSysDlgSelectColor(M_WINDOW(p), &col))
				{
					_change_color(p, col);

					mColorButtonSetColor(p->colbtt, col);
				}
			}
			break;

		//色ボタン
		case WID_BTT_SETCOLOR:
			if(type == MCOLORBUTTON_N_PRESS)
				_change_color(p, ev->notify.param1);
			break;
	}
}

/** イベントハンドラ */

static int _event_handle(mWidget *wg,mEvent *ev)
{
	MainWindow *p = (MainWindow *)wg;

	switch(ev->type)
	{
		//通知
		case MEVENT_NOTIFY:
			_event_notify(p, ev);
			break;
	
		//コマンド
		case MEVENT_COMMAND:
			_event_command(p, ev->cmd.id);
			break;

		//終了
		case MEVENT_CLOSE:
			mAppQuit();
			break;
	}

	return 1;
}

/** D&D ハンドラ */

static int _dnd_handle(mWidget *wg,char **files)
{
	_load_file((MainWindow *)wg, files[0]);

	return 1;
}


//-------------


/** 色リストのアイテム描画ハンドラ */

static void _listitem_draw_handle(mPixbuf *pixbuf,mListViewItem *pi,mListViewItemDraw *info)
{
	mFont *font = mWidgetGetFont(info->widget);
	int x,h,flags;

	h = info->box.h;
	flags = info->flags;

	//テキスト

	mFontDrawText(font, pixbuf,
		info->box.x, info->box.y + (h - font->height) / 2,
		pi->text, -1,
		(flags & MLISTVIEWITEMDRAW_F_FOCUS_ITEM)? MSYSCOL_RGB(TEXT_SELECT): MSYSCOL_RGB(TEXT));

	//色

	x = info->box.x + info->box.w - LISTITEM_COLORBOX_W;

	mPixbufBox(pixbuf, x, info->box.y + 1, LISTITEM_COLORBOX_W, h - 2, 0);

	mPixbufFillBox(pixbuf, x + 1, info->box.y + 2, LISTITEM_COLORBOX_W - 2, h - 4,
		SYSCOL_PIX(pi->param));
}

/** mDockWidget 内容作成 */

static mWidget *_dockwg_create_handle(mDockWidget *dock,int id,mWidget *parent)
{
	mWidget *ct,*wg;

	ct = mContainerCreate(parent, MCONTAINER_TYPE_VERT, 0, 4, MLF_EXPAND_WH);

	mContainerSetPadding_one(M_CONTAINER(ct), 6);

	wg = (mWidget *)mLabelCreate(ct, 0, 0, 0, "label");
	wg->initW = 120;

	return ct;
}

/** mDockWidget 配置 */

static int _dockwg_arrange_handle(mDockWidget *dock,int id1,int id2)
{
	return (id1 < id2)? -1: 1;
}


//===========================
// ウィンドウ作成
//===========================


/** アクセラレータ作成 */

static void _create_accelerator(MainWindow *p)
{
	mAccelerator *acc;

	acc = mAcceleratorNew();

	p->win.accelerator = acc;

	mAcceleratorSetDefaultWidget(acc, M_WIDGET(p));

	mAcceleratorAdd(acc, MENUID_FILE_OPEN, MACCKEY_CTRL + 'O', NULL);
	mAcceleratorAdd(acc, MENUID_FILE_SAVE, MACCKEY_CTRL + 'S', NULL);
	mAcceleratorAdd(acc, MENUID_FILE_RESET, MACCKEY_CTRL + 'R', NULL);
}

/** メニュー作成 */

static void _create_menu(MainWindow *p)
{
	mMenuBar *bar;
	mMenu *menu_top,*menu;

	bar = mMenuBarNew(0, M_WIDGET(p), 0);

	p->win.menubar = bar;

	//

	menu_top = mMenuNew();
	mMenuBarSetMenu(bar, menu_top);

	//file

	menu = mMenuNew();

	mMenuAddNormal(menu, MENUID_FILE_OPEN, "&Open...", MACCKEY_CTRL + 'O', 0);
	mMenuAddNormal(menu, MENUID_FILE_SAVE, "&Save...", MACCKEY_CTRL + 'S', 0);
	mMenuAddSep(menu);
	mMenuAddNormal(menu, MENUID_FILE_RESET, "&Reset", MACCKEY_CTRL + 'R', 0);
	mMenuAddSep(menu);
	mMenuAddText_static(menu, MENUID_FILE_EXIT, "&Exit");

	mMenuAddSubmenu(menu_top, MENUID_TOP_FILE, "&File", menu);

	//test

	menu = mMenuNew();

	mMenuAddNormal(menu, MENUID_TEST_DISABLE, "Disable", 0, MMENUITEM_F_DISABLE);
	mMenuAddNormal(menu, MENUID_TEST_CHECK, "Check", 0, MMENUITEM_F_CHECKED | MMENUITEM_F_AUTOCHECK);
	mMenuAddSep(menu);
	mMenuAddNormal(menu, MENUID_TEST_RADIO1, "Radio1", 0, MMENUITEM_F_CHECKED | MMENUITEM_F_RADIO | MMENUITEM_F_AUTOCHECK);
	mMenuAddNormal(menu, MENUID_TEST_RADIO2, "Radio2", 0, MMENUITEM_F_RADIO | MMENUITEM_F_AUTOCHECK);

	mMenuAddSubmenu(menu_top, MENUID_TOP_TEST, "&Test", menu);

	//help

	menu = mMenuNew();

	mMenuAddText_static(menu, MENUID_HELP_ABOUT, "&About...");

	mMenuAddSubmenu(menu_top, MENUID_TOP_HELP, "&Help", menu);
}

/** アイコンボタン作成 */

static void _create_iconbuttons(mWidget *parent)
{
	mIconButtons *ib;

	ib = mIconButtonsNew(0, parent,
		MICONBUTTONS_S_TOOLTIP | MICONBUTTONS_S_SEP_BOTTOM | MICONBUTTONS_S_DESTROY_IMAGELIST);

	mIconButtonsSetImageList(ib,
		mImageListLoadPNG_fromBuf(g_png_iconbtt, sizeof(g_png_iconbtt), 16, 0x00ff00));

	mIconButtonsSetTooltipTrGroup(ib, 0);

	mIconButtonsAdd(ib, 0, 0, 0, 0);
	mIconButtonsAdd(ib, 1, 1, 1, MICONBUTTONS_ITEMF_CHECKBUTTON | MICONBUTTONS_ITEMF_CHECKED);
	mIconButtonsAddSep(ib);
	mIconButtonsAdd(ib, 2, 2, 2, MICONBUTTONS_ITEMF_CHECKGROUP | MICONBUTTONS_ITEMF_CHECKED);
	mIconButtonsAdd(ib, 3, 3, 3, MICONBUTTONS_ITEMF_CHECKGROUP);
	mIconButtonsAddSep(ib);
	mIconButtonsAdd(ib, 4, 4, 4, MICONBUTTONS_ITEMF_DISABLE);
	mIconButtonsAdd(ib, 5, 5, 5, MICONBUTTONS_ITEMF_DROPDOWN);

	mIconButtonsCalcHintSize(ib);
}

/** mDockWidget 作成 */

static void _create_dockwidget(MainWindow *p,mWidget *parent)
{
	mDockWidget *dock;
	mDockWidgetState st;
	int i;
	char m[16];

	st.flags = MDOCKWIDGET_F_EXIST | MDOCKWIDGET_F_VISIBLE;
	st.dockH = 120;
	st.boxWin.x = st.boxWin.y = 0;
	st.boxWin.w = st.boxWin.h = 200;

	for(i = 0; i < DOCKWG_NUM; i++)
	{
		dock = p->dockwg[i] = mDockWidgetNew(i, MDOCKWIDGET_S_ALLBUTTON);

		mDockWidgetSetCallback_create(dock, _dockwg_create_handle);
		mDockWidgetSetCallback_arrange(dock, _dockwg_arrange_handle);

		mDockWidgetSetWindowInfo(dock, M_WINDOW(parent->toplevel),
			MWINDOW_S_NORMAL | MWINDOW_S_TABMOVE | MWINDOW_S_OWNER);

		mDockWidgetSetDockParent(dock, parent);

		snprintf(m, 16, "dock%d", i + 1);
		mDockWidgetSetTitle(dock, m);

		mDockWidgetSetState(dock, &st);

		mDockWidgetCreateWidget(dock);
	}
}

/** サンプルウィジェット作成 */

static void _create_sample_widgets(MainWindow *p,mWidget *parent)
{
	mWidget *ct_topv,*ct_h,*ct_v,*ct,*wg;
	mGroupBox *gb;
	mComboBox *cb;
	mTab *tab;
	char m[16];
	int i;

	ct_topv = mContainerCreate(parent, MCONTAINER_TYPE_VERT, 0, 10, MLF_EXPAND_WH);

	//アイコンボタン

	_create_iconbuttons(ct_topv);

	//-----

	ct_h = mContainerCreate(ct_topv, MCONTAINER_TYPE_HORZ, 0, 10, MLF_EXPAND_WH);

	//======== 左

	ct_v = mContainerCreate(ct_h, MCONTAINER_TYPE_VERT, 0, 10, MLF_EXPAND_H);

	gb = mGroupBoxCreate(ct_v, 0, 0, 0, "group1");
	gb->ct.sepW = 9;

	//label + button

	ct = mContainerCreate(M_WIDGET(gb), MCONTAINER_TYPE_HORZ, 0, 5, 0);

	mLabelCreate(ct, 0, MLF_MIDDLE, 0, "label");
	mButtonCreate(ct, 0, 0, 0, 0, "Button");

	//checkbutton

	mCheckButtonCreate(M_WIDGET(gb), 0, 0, 0, 0, "check button", 1);

	//radio button (disable)

	wg = (mWidget *)mCheckButtonCreate(M_WIDGET(gb), 0, MCHECKBUTTON_S_RADIO, 0, 0, "radio button", 1);
	wg->fState &= ~MWIDGET_STATE_ENABLED;

	//combobox

	cb = mComboBoxCreate(M_WIDGET(gb), 0, 0, 0, 0);

	for(i = 1; i <= 6; i++)
	{
		snprintf(m, 16, "item%d", i);
		mComboBoxAddItem(cb, m, 0);
	}

	mComboBoxSetWidthAuto(cb);
	mComboBoxSetSel_index(cb, 0);

	//lineedit

	wg = (mWidget *)mLineEditCreate(M_WIDGET(gb), 0, 0, MLF_EXPAND_W, 0);
	mLineEditSetText(M_LINEEDIT(wg), "line edit");

	//tab

	tab = mTabCreate(M_WIDGET(gb), 0, MTAB_S_TOP | MTAB_S_HAVE_SEP, MLF_EXPAND_W, 0);

	for(i = 1; i <= 3; i++)
	{
		snprintf(m, 16, "tab%d", i);
		mTabAddItem(tab, m, -1, 0);
	}

	mTabSetSel_index(tab, 0);

	//slider bar

	wg = (mWidget *)mSliderBarCreate(M_WIDGET(gb), 0, 0, MLF_EXPAND_W, 0);
	mSliderBarSetStatus(M_SLIDERBAR(wg), 0, 100, 50);

	//progress bar

	wg = (mWidget *)mProgressBarCreate(M_WIDGET(gb), 0,
		MPROGRESSBAR_S_FRAME | MPROGRESSBAR_S_TEXT_PERS, MLF_EXPAND_W, 0);

	mProgressBarSetStatus(M_PROGRESSBAR(wg), 0, 100, 50);

	//scroll bar

	wg = (mWidget *)mScrollBarNew(0, M_WIDGET(gb), MSCROLLBAR_S_HORZ);
	wg->fLayout = MLF_EXPAND_W;

	mScrollBarSetStatus(M_SCROLLBAR(wg), 0, 100, 20);

	//========= 右

	ct = mContainerCreate(ct_h, MCONTAINER_TYPE_VERT, 0, 0, MLF_EXPAND_WH);

	_create_dockwidget(p, ct);
}

/** ウィジェット作成 */

static void _create_widgets(MainWindow *p)
{
	mWidget *ct_toph,*ct,*ct2;
	mListView *list;
	mListViewItem *pi;
	mFont *font;
	int i,w,maxw;

	ct_toph = mContainerCreate(M_WIDGET(p), MCONTAINER_TYPE_HORZ, 0, 15, MLF_EXPAND_WH);

	mContainerSetPadding_one(M_CONTAINER(ct_toph), 8);

	//------ リスト

	ct = mContainerCreate(ct_toph, MCONTAINER_TYPE_VERT, 0, 6, MLF_EXPAND_H);

	p->list = list = mListViewNew(0, ct, 0, MSCROLLVIEW_S_VERT | MSCROLLVIEW_S_FRAME);

	list->wg.id = WID_LIST_COLOR;
	list->wg.fLayout = MLF_EXPAND_H;

	font = mWidgetGetFont(M_WIDGET(list));

	i = font->height + 2;
	if(i < 16) i = 16;
	list->lv.itemHeight = i;

	maxw = 0;

	for(i = 0; i < MSYSCOL_NUM - 1; i++)
	{
		pi = mListViewAddItem(list, g_color_name[i], -1, MLISTVIEW_ITEM_F_STATIC_TEXT, i);

		pi->draw = _listitem_draw_handle;

		w = mFontGetTextWidth(font, g_color_name[i], -1);
		if(maxw < w) maxw = w;
	}

	p->selitem = mListViewSetFocusItem_index(list, 0);

	list->wg.hintOverW = mListViewCalcWidgetWidth(list, maxw) + LISTITEM_COLORBOX_W + LISTITEM_COLORBOX_SEP;
	list->wg.initH = mListViewGetItemNormalHeight(list) * (MSYSCOL_NUM - 1) + 4;

	//ボタン

	ct2 = mContainerCreate(ct, MCONTAINER_TYPE_HORZ, 0, 5, MLF_EXPAND_W);

	mLabelCreate(ct2, 0, MLF_MIDDLE, 0, "Color:");

	p->colbtt = mColorButtonCreate(ct2, WID_BTT_SETCOLOR, MCOLORBUTTON_S_DIALOG, MLF_EXPAND_W, 0, SYSCOL_RGB(0));

	//------ サンプル

	_create_sample_widgets(p, ct_toph);
}

/** 破棄ハンドラ */

static void _destroy_handle(mWidget *wg)
{
	MainWindow *p = (MainWindow *)wg;
	int i;

	mAcceleratorDestroy(p->win.accelerator);

	for(i = 0; i < DOCKWG_NUM; i++)
		mDockWidgetDestroy(p->dockwg[i]);

	mStrFree(&p->strOpenDir);
	mStrFree(&p->strSaveDir);
}

/** メインウィンドウ作成 */

void createMainWindow()
{
	MainWindow *p;

	//ウィンドウ
	
	p = (MainWindow *)mWindowNew(sizeof(MainWindow), NULL,
		MWINDOW_S_NORMAL | MWINDOW_S_TABMOVE | MWINDOW_S_ENABLE_DND);
	if(!p) return;
	
	p->wg.event = _event_handle;
	p->wg.destroy = _destroy_handle;
	p->wg.onDND = _dnd_handle;
	p->wg.fState |= MWIDGET_STATE_ENABLE_DROP;

	mWindowSetTitle(M_WINDOW(p), "mThemeEditor");

	mWindowEnableDND(M_WINDOW(p));

	//作成

	_create_menu(p);

	_create_accelerator(p);

	_create_widgets(p);

	//表示

	mWindowMoveResizeShow_hintSize(M_WINDOW(p));
}


//===========================
// メイン
//===========================


/** メイン */

int main(int argc,char **argv)
{
	if(mAppInit(&argc, argv)) return 1;

	mAppLoadTranslation(g_deftransdat, NULL, NULL);

	//初期化

	createMainWindow();

	//実行

	mAppRun();

	//終了

	mAppEnd();

	return 0;
}
